/*
 * Written by Solar Designer <solar at openwall.com> in 2014.
 * No copyright is claimed, and the software is hereby placed in the public
 * domain.  In case this attempt to disclaim copyright and place the software
 * in the public domain is deemed null and void, then the software is
 * Copyright (c) 2014 Solar Designer and it is hereby released to the
 * general public under the following terms:
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted.
 *
 * There's ABSOLUTELY NO WARRANTY, express or implied.
 */

#include <stdio.h>
#include <string.h>

#include "crypt_blowfish/ow-crypt.h"

/* Reference, same prototype as in OpenBSD's crypt.c */
extern char *bcrypt(const char *, const char *);

static char *patch(char *dst, int dst_size, const char *src, char version)
{
	dst[0] = 0;
	strncat(dst, src, dst_size - 1);
	dst[2] = version;
	return dst;
}

/* Unit under test */
static char *bcrypt_uut(const char *key, const char *setting)
{
	static void *data = NULL;
	static int size = 0;
	char *retval;
#if 0
/* Emulate future crypt_blowfish's support for 'b' as alias for 'y' */
	char buf[128];
	if (!strncmp(setting, "$2b$", 4))
		setting = patch(buf, sizeof(buf), setting, 'y');
#else
/* The future is now */
#endif
/*
 * Use a crypt_blowfish specific function name here to be sure we're calling
 * the implementation being tested rather than the underlying system's crypt().
 */
	retval = crypt_ra(key, setting, &data, &size);
#if 0
/* Emulate future crypt_blowfish's support for 'b' as alias for 'y' */
	if (retval && !strncmp(retval, "$2y$", 4))
		retval[2] = 'b';
#else
/* The future is now */
#endif
	return retval;
}

/*
 * Test vectors copied from crypt_blowfish/wrapper.c #ifdef TEST section.
 */
static const char * const tests[][4] = {
	{"$2a$05$CCCCCCCCCCCCCCCCCCCCC.E5YPO9kmyuRGyh0XouQYb4YMJKvyOeW",
		"U*U"},
	{"$2a$05$CCCCCCCCCCCCCCCCCCCCC.VGOzA784oUp/Z0DY336zx7pLYAy0lwK",
		"U*U*"},
	{"$2a$05$XXXXXXXXXXXXXXXXXXXXXOAcXxm9kjPGEMsLznoKqmqw7tc8WCx4a",
		"U*U*U"},
	{"$2a$05$abcdefghijklmnopqrstuu5s2v8.iXieOjg/.AySBTTZIIVFJeBui",
		"0123456789abcdefghijklmnopqrstuvwxyz"
		"ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
		"chars after 72 are ignored"},
	{"$2x$05$/OK.fbVrR/bpIqNJ5ianF.CE5elHaaO4EbggVDjb8P19RukzXSM3e",
		"\xa3"},
	{"$2x$05$/OK.fbVrR/bpIqNJ5ianF.CE5elHaaO4EbggVDjb8P19RukzXSM3e",
		"\xff\xff\xa3"},
	{"$2y$05$/OK.fbVrR/bpIqNJ5ianF.CE5elHaaO4EbggVDjb8P19RukzXSM3e",
		"\xff\xff\xa3"},
	{"$2a$05$/OK.fbVrR/bpIqNJ5ianF.nqd1wy.pTMdcvrRWxyiGL2eMz.2a85.",
		"\xff\xff\xa3", NULL, "F"},
	{"$2b$05$/OK.fbVrR/bpIqNJ5ianF.CE5elHaaO4EbggVDjb8P19RukzXSM3e",
		"\xff\xff\xa3"},
	{"$2y$05$/OK.fbVrR/bpIqNJ5ianF.Sa7shbm4.OzKpvFnX1pQLmQW96oUlCq",
		"\xa3"},
	{"$2a$05$/OK.fbVrR/bpIqNJ5ianF.Sa7shbm4.OzKpvFnX1pQLmQW96oUlCq",
		"\xa3"},
	{"$2b$05$/OK.fbVrR/bpIqNJ5ianF.Sa7shbm4.OzKpvFnX1pQLmQW96oUlCq",
		"\xa3"},
	{"$2x$05$/OK.fbVrR/bpIqNJ5ianF.o./n25XVfn6oAPaUvHe.Csk4zRfsYPi",
		"1\xa3" "345"},
	{"$2x$05$/OK.fbVrR/bpIqNJ5ianF.o./n25XVfn6oAPaUvHe.Csk4zRfsYPi",
		"\xff\xa3" "345"},
	{"$2x$05$/OK.fbVrR/bpIqNJ5ianF.o./n25XVfn6oAPaUvHe.Csk4zRfsYPi",
		"\xff\xa3" "34" "\xff\xff\xff\xa3" "345"},
	{"$2y$05$/OK.fbVrR/bpIqNJ5ianF.o./n25XVfn6oAPaUvHe.Csk4zRfsYPi",
		"\xff\xa3" "34" "\xff\xff\xff\xa3" "345"},
	{"$2a$05$/OK.fbVrR/bpIqNJ5ianF.ZC1JEJ8Z4gPfpe1JOr/oyPXTWl9EFd.",
		"\xff\xa3" "34" "\xff\xff\xff\xa3" "345", NULL, "F"},
	{"$2y$05$/OK.fbVrR/bpIqNJ5ianF.nRht2l/HRhr6zmCp9vYUvvsqynflf9e",
		"\xff\xa3" "345"},
	{"$2a$05$/OK.fbVrR/bpIqNJ5ianF.nRht2l/HRhr6zmCp9vYUvvsqynflf9e",
		"\xff\xa3" "345"},
	{"$2a$05$/OK.fbVrR/bpIqNJ5ianF.6IflQkJytoRVc1yuaNtHfiuq.FRlSIS",
		"\xa3" "ab"},
	{"$2x$05$/OK.fbVrR/bpIqNJ5ianF.6IflQkJytoRVc1yuaNtHfiuq.FRlSIS",
		"\xa3" "ab"},
	{"$2y$05$/OK.fbVrR/bpIqNJ5ianF.6IflQkJytoRVc1yuaNtHfiuq.FRlSIS",
		"\xa3" "ab"},
	{"$2x$05$6bNw2HLQYeqHYyBfLMsv/OiwqTymGIGzFsA4hOTWebfehXHNprcAS",
		"\xd1\x91"},
	{"$2x$05$6bNw2HLQYeqHYyBfLMsv/O9LIGgn8OMzuDoHfof8AQimSGfcSWxnS",
		"\xd0\xc1\xd2\xcf\xcc\xd8"},
	{"$2a$05$/OK.fbVrR/bpIqNJ5ianF.swQOIzjOiJ9GHEPuhEkvqrUyvWhEMx6",
		"\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa"
		"\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa"
		"\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa"
		"\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa"
		"\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa"
		"\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa"
		"chars after 72 are ignored as usual"},
	{"$2a$05$/OK.fbVrR/bpIqNJ5ianF.R9xrDjiycxMbQE2bp.vgqlYpW5wx2yy",
		"\xaa\x55\xaa\x55\xaa\x55\xaa\x55\xaa\x55\xaa\x55"
		"\xaa\x55\xaa\x55\xaa\x55\xaa\x55\xaa\x55\xaa\x55"
		"\xaa\x55\xaa\x55\xaa\x55\xaa\x55\xaa\x55\xaa\x55"
		"\xaa\x55\xaa\x55\xaa\x55\xaa\x55\xaa\x55\xaa\x55"
		"\xaa\x55\xaa\x55\xaa\x55\xaa\x55\xaa\x55\xaa\x55"
		"\xaa\x55\xaa\x55\xaa\x55\xaa\x55\xaa\x55\xaa\x55"},
	{"$2a$05$/OK.fbVrR/bpIqNJ5ianF.9tQZzcJfm3uj2NvJ/n5xkhpqLrMpWCe",
		"\x55\xaa\xff\x55\xaa\xff\x55\xaa\xff\x55\xaa\xff"
		"\x55\xaa\xff\x55\xaa\xff\x55\xaa\xff\x55\xaa\xff"
		"\x55\xaa\xff\x55\xaa\xff\x55\xaa\xff\x55\xaa\xff"
		"\x55\xaa\xff\x55\xaa\xff\x55\xaa\xff\x55\xaa\xff"
		"\x55\xaa\xff\x55\xaa\xff\x55\xaa\xff\x55\xaa\xff"
		"\x55\xaa\xff\x55\xaa\xff\x55\xaa\xff\x55\xaa\xff"},
	{"$2a$05$CCCCCCCCCCCCCCCCCCCCC.7uG0VCzI2bS7j6ymqJi9CdcdxiRTWNy",
		""},
	{"*0", "", "$2a$03$CCCCCCCCCCCCCCCCCCCCC."},
	{"*0", "", "$2a$32$CCCCCCCCCCCCCCCCCCCCC."},
	{"*0", "", "$2c$05$CCCCCCCCCCCCCCCCCCCCC."},
	{"*0", "", "$2z$05$CCCCCCCCCCCCCCCCCCCCC."},
	{"*0", "", "$2`$05$CCCCCCCCCCCCCCCCCCCCC."},
	{"*0", "", "$2{$05$CCCCCCCCCCCCCCCCCCCCC."},
	{"*1", "", "*0"},

	{"*0", "", "$2b$00$CCCCCCCCCCCCCCCCCCCCC."},
	{"*0", "", "$2b$01$CCCCCCCCCCCCCCCCCCCCC."},
	{"*0", "", "$2b$02$CCCCCCCCCCCCCCCCCCCCC."},
	{"*0", "", "$2b$03$CCCCCCCCCCCCCCCCCCCCC."},
	{"*0", "", "$2b$32$CCCCCCCCCCCCCCCCCCCCC."},
	{"*0", "", "$2b$33$CCCCCCCCCCCCCCCCCCCCC."},
	{"*0", "", "$2b$62$CCCCCCCCCCCCCCCCCCCCC."},
	{"*0", "", "$2b$63$CCCCCCCCCCCCCCCCCCCCC."},
	{"*0", "", "$2b$64$CCCCCCCCCCCCCCCCCCCCC."},
	{"*0", "", "$2b$65$CCCCCCCCCCCCCCCCCCCCC."},
	{"*0", "", "$2b$99$CCCCCCCCCCCCCCCCCCCCC."},
	{"*0", "", "$2b$9a$CCCCCCCCCCCCCCCCCCCCC."},
	{"*0", "", "$2b$a9$CCCCCCCCCCCCCCCCCCCCC."},
	{"*0", "", "$2b$1\x80$CCCCCCCCCCCCCCCCCCCCC."},
	{"*0", "", "$2b$\x80" "1$CCCCCCCCCCCCCCCCCCCCC."},

	{NULL}
};

int main(void)
{
	unsigned int i;

	for (i = 0; tests[i][0]; i++) {
		const char *hash = tests[i][0];
		const char *key = tests[i][1];
		const char *setting = tests[i][2] ? tests[i][2] : hash;
		const char *ref, *uut;
		int ref_ok, uut_ok;

		if (!strncmp(setting, "$2y$", 4)) {
			static char buf1[128], buf2[128];
			setting = patch(buf1, sizeof(buf1), setting, 'b');
			hash = patch(buf2, sizeof(buf2), hash, 'b');
		}

		ref = bcrypt(key, setting);
		uut = bcrypt_uut(key, setting);

/*
 * In this program, we're using a crypt_blowfish function that is defined to
 * return NULL on error, not the crypt() wrapper with its retval magic.  The
 * magic should be tested via crypt_blowfish's own "make check".
 */
		if (hash[0] == '*')
			hash = ":"; /* match reference */
		if (!uut)
			uut = ":"; /* match reference */

		uut_ok = !strcmp(uut, hash);
		ref_ok = ref && !strcmp(ref, hash);

/*
 * Reference implementation will usually "fail" on crypt_blowfish's collision
 * countermeasure tests (marked in the tests[] above by setting the fourth
 * string pointer to non-empty).  Let's actually expect such "failures".
 */
		if (!strncmp(setting, "$2a$", 4) && tests[i][3])
			ref_ok = !ref_ok;

		if (!strncmp(setting, "$2x$", 4))
			ref_ok = 1;

		if (!uut_ok || !ref_ok)
			printf("key: \"%s\"\nset: \"%s\"\n"
			    "exp: \"%s\"\nref: \"%s\"\nuut: \"%s\"\n",
			    key, setting, hash, ref, uut);
	}

	for (i = 0; i < 1500; i++) {
		const char *setting = "$2b$04$XXXXXXXXXXXXXXXXXXXXXO";
		char key[700];
		const char *ref, *uut;
		unsigned int j, len = i % sizeof(key);

		for (j = 0; j < len; j++)
			key[j] = (i + j) % 255 + 1;
		key[j] = 0;

		ref = bcrypt(key, setting);
		uut = bcrypt_uut(key, setting);

		if (strcmp(ref, uut))
			printf("key: \"%s\"\nset: \"%s\"\n"
			    "ref: \"%s\"\nuut: \"%s\"\n",
			    key, setting, ref, uut);
	}

	for (i = 0; i < 1500; i++) {
		char setting[128];
		const char * const key = tests[0][1];
#if 1
		unsigned int j = i % 60;
#else
/*
 * Reference implementation assumes that setting starts with "$2", so we don't
 * want info on this known issue (or non-issue, depending on context).
 */
		unsigned int j = i % 58 + 2;
#endif
		unsigned char c = i % 255 + 1;
		const char *ref, *uut;

		if (j == 4 && (c == '2' || c == '3'))
			c = '4';

		setting[0] = 0;
		strncat(setting, tests[0][0], sizeof(setting) - 1);
		setting[j] = c;

		ref = bcrypt(key, setting);
		uut = bcrypt_uut(key, setting);

		if (!uut)
			uut = ":"; /* match reference */

		if (strcmp(ref, uut))
			printf("set: \"%s\"\nref: \"%s\"\nuut: \"%s\"\n",
			    setting, ref, uut);
	}

	return 0;
}
